/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_utils_hh
#define pplite_utils_hh 1

#include "globals.hh"

#include <iterator>
#include <vector>
#include <cassert>

namespace pplite {

// This helper function returns a dim_type value:
// used to avoid warnings about signed/unsigned comparisons.
template <typename Row>
inline dim_type
num_rows(const std::vector<Row>& rows) { return rows.size(); }

template <typename Row, typename Iter>
inline void
erase_using_sorted_indices(std::vector<Row>& rows,
                           Iter first, Iter last) {
  if (first == last)
    return;

  // `idx_unused' and `idx' start with value `*first' (instead of `0'),
  // because the loop would just increment them in the preceding iterations.
  using Index = typename std::iterator_traits<Iter>::value_type;
  const Index idx_end = rows.size();
  auto idx_unused = *first;
  assert(0 <= idx_unused);
  if (idx_unused >= idx_end)
    return;
  auto idx = idx_unused + 1;
  ++first;
  while (first != last && idx != idx_end) {
    auto new_idx = *first;
    assert(idx_unused < idx && idx <= new_idx);
    if (idx == new_idx) {
      // Don't swap row idx.
      ++first;
      ++idx;
    } else {
      // Keep (i.e., swap to its correct place) row idx.
      using std::swap;
      swap(rows[idx_unused], rows[idx]);
      ++idx_unused;
      ++idx;
    }
  }
  // Move up the remaining rows, if any.
  for ( ; idx < idx_end; ++idx) {
    using std::swap;
    swap(rows[idx_unused], rows[idx]);
    ++idx_unused;
  }
  // The trailing rows have to be erased.
  rows.resize(idx_unused);
}

template <typename Row, typename Indices>
inline void
erase_using_sorted_indices(std::vector<Row>& rows, const Indices& indices) {
  erase_using_sorted_indices(rows, std::begin(indices), std::end(indices));
}

inline size_t
hash_size(size_t val) {
  std::hash<size_t> hash_fn;
  return hash_fn(val);
}

/* Boost code (GPL compatible) */
inline void
hash_combine(size_t& lhs, size_t rhs) {
  lhs ^= rhs + 0x9e3779b9 + (lhs << 6) + (lhs >> 2);
}

} // namespace pplite

#endif // !defined(pplite_utils_hh)
