/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite_test.hh"

using pplite::IO_Operators::operator<<;

bool
test01() {
  Var x1(0);
  Var x2(1);

  Poly ph(2);
  ph.add_con(-2*x1-x2 >= -5);
  ph.add_con(4*x1-4*x2 >= -5);
  ph.add_con(x1 >= 0);
  ph.add_con(x2 >= 0);

  print_cons(ph, "*** ph ***");

  Linear_Expr le = x1 - 2*x2;
  Integer num;
  Integer den;
  bool included;
  Gen g(point());
  bool ok = ph.max(le, 0, num, den, included, g)
    && num == 5 && den == 2 && included
    && g.is_point()
    && g.coeff(x1) == 5 && g.coeff(x2) == 0
    && g.divisor() == 2;

  nout << (included ? "maximum" : "supremum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  if (!ok)
    return false;

  ok = ph.min(le, 0, num, den, included, g)
    && num == -15 && den == 4 && included
    && g.is_point()
    && g.coeff(x1) == 5 && g.coeff(x2) == 10
    && g.divisor() == 4;

  nout << (included ? "minimum" : "infimum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  return ok;
}

bool
test02() {
  Var x1(0);
  Var x2(1);
  Var x3(2);

  Poly ph(3);
  ph.add_con(-x1-x2-x3 >= -100);
  ph.add_con(-10*x1-4*x2-5*x3 >= -600);
  ph.add_con(-x1-x2-3*x3 >= -150);
  ph.add_con(x1 >= 0);
  ph.add_con(x2 >= 0);
  ph.add_con(x3 >= 0);

  print_cons(ph, "*** ph ***");

  Linear_Expr le = -10*x1 - 6*x2 - 4*x3;
  Integer inhomo = 4;

  Integer num;
  Integer den;
  bool included;
  Gen g(point());
  bool ok = ph.max(le, inhomo, num, den, included, g)
    && num == 4 && den == 1 && included
    && g.is_point()
    && g.coeff(x1) == 0
    && g.coeff(x2) == 0
    && g.coeff(x3) == 0
    && g.divisor() == 1;

  nout << (included ? "maximum" : "supremum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  if (!ok)
    return false;

  ok = ph.min(le, inhomo, num, den, included, g)
    && num == -2188 && den == 3 && included
    && g.is_point()
    && g.coeff(x1) == 100
    && g.coeff(x2) == 200
    && g.coeff(x3) == 0
    && g.divisor() == 3;

  nout << (included ? "minimum" : "infimum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  return ok;
}

bool
test03() {
  Poly ph(0);

  print_cons(ph, "*** ph ***");

  Integer num;
  Integer den;
  bool included;
  Gen g(point());
  Linear_Expr le;
  bool ok = ph.max(le, 0, num, den, included, g)
    && num == 0 && den == 1 && included
    && g.is_point()
    && g.divisor() == 1;

  nout << (included ? "maximum" : "supremum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  if (!ok)
    return false;

  ok = ph.min(le, 0, num, den, included, g)
    && num == 0 && den == 1 && included
    && g.is_point()
    && g.divisor() == 1;

  nout << (included ? "minimum" : "infimum") << " = " << num;
  if (den != 1)
    nout << "/" << den;
  nout << " @ " << g << endl;

  return ok;
}

bool
test04() {
  Poly ph(2, Spec_Elem::EMPTY);

  print_cons(ph, "*** ph ***");

  Integer num = 0;
  Integer den = 0;
  bool included = false;
  Gen g(point());
  Linear_Expr le;
  bool ok = !ph.max(le, 0, num, den, included, g)
    && num == 0 && den == 0 && !included
    && g.is_point()
    && g.divisor() == 1;
  ok = ok && !ph.min(le, 0, num, den, included, g)
    && num == 0 && den == 0 && !included
    && g.is_point()
    && g.divisor() == 1;

  return ok;
}

BEGIN_MAIN
  DO_TEST(test01);
  DO_TEST(test02);
  DO_TEST(test03);
  DO_TEST(test04);
END_MAIN
